#!/usr/bin/env python3
#******************************************************************************
# Copyright (C) 2018 Thomas "Cakeisalie5" Touhey <thomas@touhey.fr>
# This file is part of the textoutpc project, which is MIT-licensed.
#******************************************************************************
""" Unit tests for the Python version of textout, lightscript-related funcs.
	Uses the builtin `unittest` module.
"""

import unittest
import textoutpc

# Define the tests.

__test_cases = {
	# Basic text.
	'': '',
}

# Define the tests wrapper, and define the classes.

_cnt = 0
_len = len(str(len(__test_cases)))
_templ = """\
  def test_lgsp{n:0>{l}}(self):
    self.assertEqual(textoutpc.tolightscript({i}), {r}, {i})
"""

def _wrap_test(inp, res):
	global _cnt

	_cnt += 1
	return _templ.format(n = _cnt, l = _len, i = repr(inp), r = repr(res))

exec("class TextoutLSTest(unittest.TestCase):\n  maxDiff = None\n" + \
	'\n'.join(map(lambda args: _wrap_test(*args), __test_cases.items())),
	globals())

# If run as main script, run the test function.

if __name__ == '__main__':
	unittest.main()

# End of file.



#!/usr/bin/env python3
#******************************************************************************
# Copyright (C) 2018 Thomas "Cakeisalie5" Touhey <thomas@touhey.fr>
# This file is part of the textoutpc project, which is MIT-licensed.
#******************************************************************************
""" Functions for the user.
	Really simplifies the thing.
"""

import io as _io
from .translate import Translator as _Translator

__all__ = ["version", "tohtml", "tolightscript"]

version = "0.1.1"

def tohtml(message, **tweaks):
	""" Converts textout BBcode to HTML.
		Receives a string, returns a string. """

	return _Translator(_io.StringIO(message), _io.StringIO(), 'html', \
		tweaks).process().getvalue()

def tolightscript(message, **tweaks):
	""" Converts textout BBcode to Lightscript.
		Receives a string, returns a string. """

	return "" # TODO: real thing one day

	return _Translator(_io.StringIO(message), _io.StringIO(), 'lightscript', \
		tweaks).process().getvalue()

# End of file.



#!/usr/bin/env python3
#******************************************************************************
# Copyright (C) 2018 Thomas "Cakeisalie5" Touhey <thomas@touhey.fr>
# This file is part of the textoutpc project, which is MIT-licensed.
#******************************************************************************
""" HTML/CSS like color parsing, mainly for the `[color]` tag.
	Defines the `get_color()` function which returns an rgba value.
"""

import re as _re
import math as _math

from .named import colors as _named_colors
from .sys import hls_to_rgb as _hls_to_rgb, hwb_to_rgb as _hwb_to_rgb

__all__ = ["get_color"]

_cr = _re.compile("""
	rgba?\s*\(
		\s* (?P<rgb_r>[0-9]{1,3}) \s* ([,\\s]
		\s* (?P<rgb_g>[0-9]{1,3}) \s* ([,\\s]
		\s* (?P<rgb_b>[0-9]{1,3}) \s* ([,\\s]
		\s* ((?P<rgb_a_per> ([0-9]+\.?|[0-9]*\.[0-9]+) ) \s*%
			|(?P<rgb_a_flt> (0*[01]\.?|0*\.[0-9]+) )) \s*
		)?)?)?
	\)|
	hsla?\s*\(
		\s* (?P<hsl_hue>-? ([0-9]+\.?|[0-9]*\.[0-9]+) )
			(?P<hsl_agl>deg|grad|rad|turn|) \s*[,\\s]
		\s* ((?P<hsl_sat_per> ([0-9]+\.?|[0-9]*\.[0-9]+) ) \s*%
			|(?P<hsl_sat_flt> (0*[01]\.?|0*\.[0-9]+) )) \s*[,\\s]
		\s* ((?P<hsl_lgt_per> ([0-9]+\.?|[0-9]*\.[0-9]+) ) \s*%
			|(?P<hsl_lgt_flt> (0*[01]\.?|0*\.[0-9]+) )) \s*([,\\s/]
		\s* ((?P<hsl_aph_per> ([0-9]+\.?|[0-9]*\.[0-9]+) ) \s*%
			|(?P<hsl_aph_flt> (0*[01]\.?|0*\.[0-9]+) )) \s*
		)?
	\)|
	hlsa?\s*\(
		\s* (?P<hls_hue>-? ([0-9]+\.?|[0-9]*\.[0-9]+) )
			(?P<hls_agl>deg|grad|rad|turn|) \s*[,\\s]
		\s* ((?P<hls_lgt_per> ([0-9]+\.?|[0-9]*\.[0-9]+) ) \s*%
			|(?P<hls_lgt_flt> (0*[01]\.?|0*\.[0-9]+) )) \s*[,\\s]
		\s* ((?P<hls_sat_per> ([0-9]+\.?|[0-9]*\.[0-9]+) ) \s*%
			|(?P<hls_sat_flt> (0*[01]\.?|0*\.[0-9]+) )) \s*([,\\s/]
		\s* ((?P<hls_aph_per> ([0-9]+\.?|[0-9]*\.[0-9]+) ) \s*%
			|(?P<hls_aph_flt> (0*[01]\.?|0*\.[0-9]+) )) \s*)?
	\)|
	hwb\s*\(
		\s* (?P<hwb_hue>-? ([0-9]+\.?|[0-9]*\.[0-9]+) )
			(?P<hwb_agl>deg|grad|rad|turn|) \s*[,\\s]
		\s* ((?P<hwb_wht_per> ([0-9]+\.?|[0-9]*\.[0-9]+) ) \s*%
			|(?P<hwb_wht_flt> (0*[01]\.?|0*\.[0-9]+) )) \s*[,\\s]
		\s* ((?P<hwb_blk_per> ([0-9]+\.?|[0-9]*\.[0-9]+) ) \s*%
			|(?P<hwb_blk_flt> (0*[01]\.?|0*\.[0-9]+) )) \s*([,\\s/]
		\s* ((?P<hwb_aph_per> ([0-9]+\.?|[0-9]*\.[0-9]+) ) \s*%
			|(?P<hwb_aph_flt> (0*[01]\.?|0*\.[0-9]+) )) \s*)?
	\)|
		\# (?P<hex_digits> [0-9a-f]+)
	|
		(?P<legacy_chars> [0-9a-z]+)
""", _re.VERBOSE | _re.I | _re.M)

def get_color(value):
	""" Get a color from a string.
		Returns an (r, g, b, a) color.
		Raises an exception if the color could not be read. """

	# Check if is a color name.
	value = value.strip()
	try:    value = _named_colors[value.lower()]
	except: pass

	# Initialize the alpha.
	alpha = 1.0

	# Get the match.
	match = _cr.fullmatch(value).groupdict()
	if   match['hex_digits'] or match['legacy_chars']:
		# Imitate the Netscape behaviour. Find more about this here:
		# https://stackoverflow.com/a/8333464
		#
		# I've also extended the thing as I could to introduce more
		# modern syntaxes described on the dedicated MDN page:
		# https://developer.mozilla.org/en-US/docs/Web/CSS/color_value
		#
		# First of all, depending on our source, we will act differently:
		# - if we are using the `hex_digits` source, then we use the modern
		#   behaviour and do the fancy things such as `#ABC -> #AABBCC`
		#   management and possible alpha decoding;
		# - if we are using the `legacy_chars` source, then we sanitize our
		#   input by replacing invalid characters by '0' characters (the
		#   0xFFFF limit is due to how UTF-16 was managed at the time).
		#   We shall also truncate our input to 128 characters.
		#
		# After these sanitization options, we will keep the same method as
		# for legacy color decoding. It should work and be tolerant enough…

		members = 3
		if match['hex_digits']:
			hx = match['hex_digits'].lower()

			# RGB and RGBA (3 and 4 char.) notations.
			if len(hx) in (3, 4):
				hx = hx[0:1] * 2 + hx[1:2] * 2 + hx[2:3] * 2 + hx[3:4] * 2

			# Check if there is transparency or not.
			if len(hx) % 3 != 0 and len(hx) % 4 == 0:
				members = 4

		else: # our source is `legacy_chars`
			hx = match['legacy_chars'].lower()
			hx = ''.join(c if c in '0123456789abcdef' \
				else ('0', '00')[ord(c) > 0xFFFF] for c in hx[:128])[:128]

		# First, calculate some values we're going to need.
		# `iv` is the size of the zone for a member.
		# `sz` is the size of the digits slice to take in that zone (max. 8).
		# `of` is the offset in the zone of the slice to take.

		iv = _math.ceil(len(hx) / members)
		of = iv - 8 if iv > 8 else 0
		sz = iv - of

		# Then isolate the slices using the values calculated above.
		# `gr` will be an array of 3 or 4 digit strings (depending on the
		# number of members).

		gr = list(map(lambda i: hx[i * iv + of:i * iv + iv] \
			.ljust(sz, '0'), range(members)))

		# Check how many digits we can skip at the beginning of each slice.

		pre = min(map(lambda x: len(x) - len(x.lstrip('0')), gr))
		pre = min(pre, sz - 2)

		# Then extract the values.

		it = map(lambda x: int('0' + x[pre:pre + 2], 16), gr)
		if members == 3:
			r, g, b = it
		else:
			r, g, b, alpha = it
			alpha /= 255.0
	elif match['rgb_r']:
		# Extract the values.

		r = int(match['rgb_r'])
		g = int(match['rgb_g']) if match['rgb_g'] else 0
		b = int(match['rgb_b']) if match['rgb_b'] else 0

		if   match['rgb_a_per']:
			alpha = float(match['rgb_a_per']) / 100.0
		elif match['rgb_a_flt']:
			alpha = float(match['rgb_a_flt'])
	elif match['hsl_hue'] or match['hls_hue']:
		# Extract the values.

		if match['hsl_hue']:
			hue = float(match['hsl_hue'])
			agl = match['hsl_agl']

			# Saturation.
			if match['hsl_sat_per']:
				sat = float(match['hsl_sat_per']) / 100.0
			else:
				sat = float(match['hsl_sat_flt'])
				if sat > 1.0:
					sat /= 100.0

			# Light.
			if match['hsl_lgt_per']:
				lgt = float(match['hsl_lgt_per']) / 100.0
			else:
				lgt = float(match['hsl_lgt_flt'])
				if lgt > 1.0:
					lgt /= 100.0

			# Alpha value.
			if   match['hsl_aph_per']:
				alpha = float(match['hsl_aph_per']) / 100.0
			elif match['hsl_aph_flt']:
				alpha = float(match['hsl_aph_flt'])
		else:
			hue = float(match['hls_hue'])
			agl = match['hls_agl']

			# Saturation.
			if match['hls_sat_per']:
				sat = float(match['hls_sat_per']) / 100.0
			else:
				sat = float(match['hls_sat_flt'])

			# Light.
			if match['hls_lgt_per']:
				lgt = float(match['hls_lgt_per']) / 100.0
			else:
				lgt = float(match['hls_lgt_flt'])

			# Alpha value.
			if   match['hls_aph_per']:
				alpha = float(match['hls_aph_per']) / 100.0
			elif match['hls_aph_flt']:
				alpha = float(match['hls_aph_flt'])

		# Prepare the angle.
		if   agl == 'grad':
			hue = hue * 400.0
		elif agl == 'rad':
			hue = hue / (2 * _math.pi)
		elif not agl or agl == 'deg':
			hue = hue / 360.0
		hue = hue % 1.0

		if sat > 1 or lgt > 1:
			raise Exception

		r, g, b = _hls_to_rgb(hue, lgt, sat)
		r, g, b = map(lambda x:int(round(x * 255)), (r, g, b))
	elif match['hwb_hue']:
		hue = float(match['hwb_hue'])
		agl = match['hwb_agl']

		# Prepare the angle.
		if   agl == 'grad':
			hue = hue * 400.0
		elif agl == 'rad':
			hue = hue / (2 * _math.pi)
		elif not agl or agl == 'deg':
			hue = hue / 360.0
		hue = hue % 1.0

		# Saturation.
		if match['hwb_wht_per']:
			wht = float(match['hwb_wht_per']) / 100.0
		else:
			wht = float(match['hwb_wht_flt'])

		# Light.
		if match['hwb_blk_per']:
			blk = float(match['hwb_blk_per']) / 100.0
		else:
			blk = float(match['hwb_blk_flt'])

		if wht > 1 or blk > 1:
			raise Exception
		r, g, b = _hwb_to_rgb(hue, wht, blk)
		r, g, b = map(lambda x: int(round(x * 255)), (r, g, b))

	if r < 0 or r > 255 or g < 0 or g > 255 or b < 0 or b > 255:
		raise Exception
	if alpha < 0.0 or alpha > 1.0:
		raise Exception
	alpha = round(alpha, 4)

	return (r, g, b, alpha)

# End of file.



#!/usr/bin/env python3
#******************************************************************************
# Copyright (C) 2018 Thomas "Cakeisalie5" Touhey <thomas@touhey.fr>
# This file is part of the textoutpc project, which is MIT-licensed.
#******************************************************************************

from ..tags import TextoutBlockTag as _TextoutBlockTag

__all__ = ["TextoutAlignTag"]

class TextoutAlignTag(_TextoutBlockTag):
	""" Main tag for aligning paragraphs.
		Example uses:
		
		[align=center]This text is centered horizontally.[/align]
		[justify]This text is justified.[/justify]
	"""

	aliases = ('[align]', '[center]', '[left]', '[right]', '[justify]')
	superblock = True
	notempty = True

	def prepare(self, name, value):
		align = None
		if not name: pass
		elif name == 'align' and value != None:
			align = value
		elif name[1:-1] in ('center', 'left', 'right', 'justify'):
			align = name[1:-1]

		self._align = align

	def begin_html(self):
		cl = []
		if self._align:
			cl.append('align-' + self._align)

		return '<div{}>'.format(' class="' + ' '.join(cl) + '"' if cl else '')

	def end_html(self):
		return '</div>'

# End of file.



#!/usr/bin/env python3
#******************************************************************************
# Copyright (C) 2018 Thomas "Cakeisalie5" Touhey <thomas@touhey.fr>
# This file is part of the textoutpc project, which is MIT-licensed.
#******************************************************************************

from ..tags import TextoutBlockTag as _TextoutBlockTag
import re as _re
import urllib.parse as _urlparse
from html import escape as _htmlescape

__all__ = ["TextoutVideoTag"]

_hexcode = _re.compile('[a-zA-Z0-9_]+')
_numcode = _re.compile('^/[0-9]+$')
_dailypath = _re.compile('^/video/([a-z0-9]+)$')

class TextoutVideoTag(_TextoutBlockTag):
	""" The video tag, puts a preview of the video whose URL is given.
		Only a few 'big' services are supported for now.
		Example uses:
		
		[video]video_url[/video]
		[video tiny]video_url[/video tiny]
		[video]https://www.youtube.com/watch?v=yhXpV8hRKxQ[/video]
	"""

	aliases = ('[video]', '[video tiny]')
	raw = True

	def prepare(self, name, value):
		""" Prepare the video tag. """

		self._sizeclass = "video-tiny" if "tiny" in name \
			else "video-medium"
		self._center = False

	def _getvideo(self, url):
		""" Try to get the video type for preprocessing. """

		url = _urlparse.urlparse(url)
		if not url.scheme in ('http', 'https'):
			raise Exception

		if   url.netloc == "youtu.be":
			self._id = url.path[1:]
			if not _hexcode.match(self._id):
				raise Exception
			self._type = "youtube"
		elif url.netloc in ('youtube.com', 'www.youtube.com'):
			if url.path != '/watch':
				raise Exception
			self._id = _urlparse.parse_qs(url.query)['v'][0]
			if not _hexcode.fullmatch(self._id):
				raise Exception
			self._type = "youtube"
		elif url.netloc in ('dailymotion.com', 'www.dailymotion.com'):
			self._code = _dailypath.match(url.path).groups()[0]
			self._type = "dailymotion"
		elif url.netloc in ('vimeo.com', 'www.vimeo.com'):
			self._code = url.path[1:]
			if not _numcode.match(self._code):
				raise Exception
			self._type = "vimeo"
		else:
			raise Exception

	def preprocess(self, content):
		self._url = content

		try:
			self._getvideo(content)
		except:
			url = _urlparse.urlparse(content)
			if not url.scheme in ('http', 'https'):
				raise Exception("No allowed prefix!")
			self._type = None

	def content_html(self):
		""" Produce the embed code for the given type. """

		if not self._type:
			url = _htmlescape(self._url)
			return '<p><a href="{}">{}</a></p>'.format(url, url)

		code = '<div class="video-wrapper {}{}">'.format(self._sizeclass,
			" video-center" if self._center else "")

		if   self._type == "youtube":
			code += '<iframe ' \
				'src="https://www.youtube.com/embed/{}" frameborder="0" ' \
				'allowfullscreen></iframe>'.format(self._id)
		elif self._type == "dailymotion":
			code += '<iframe frameborder="0" ' \
				'src="https://www.dailymotion.com/embed/video/{}">' \
				'</iframe>'.format(self._code)
		elif self._type == "vimeo":
			code += '<iframe src="https://player.vimeo.com/video/{}' \
				'?title=0&byline=0&portrait=0" frameborder="0" ' \
				'webkitAllowFullScreen allowFullScreen>' \
				'</iframe>'.format(self._code)

		return code + '</div>'

	def content_lightscript(self):
		url = self._url.replace('[', '%5B').replace(']', '%5D')
		return '[[image:{}]]'.format(url)

# End of file.



#!/usr/bin/env python3
#******************************************************************************
# Copyright (C) 2018 Thomas "Cakeisalie5" Touhey <thomas@touhey.fr>
# This file is part of the textoutpc project, which is MIT-licensed.
#******************************************************************************

from ..tags import TextoutBlockTag as _TextoutBlockTag

__all__ = ["TextoutProgressTag"]

class TextoutProgressTag(_TextoutBlockTag):
	""" Progress tag, used to display the progress on anything.
		Usage:
		
		[progress=50]My great progress bar[/progress]
		[progress=100][/progress] """

	aliases = ('[progress]',)
	raw = True

	def prepare(self, name, value):
		self._val = int(value)
		if self._val < 0 or self._val > 100:
			raise Exception("progress value should be between 0 and 100 incl.")

	def begin_html(self):
		return '<div>'

	def end_html(self):
		return '' \
		'<div style="background-color: white; border: 1px solid black; ' \
			'width: 50%; margin-top: 2px; text-align: left;">' \
		'<div style="background-color: #FF3E28; color: black; ' \
			'font-weight: bold; max-width: 100%; width: {}%;' \
			'height: 18px;">   {}%' \
		'</div></div></div>'.format(self._val, self._val)

# End of file.



#!/usr/bin/env python3
#******************************************************************************
# Copyright (C) 2018 Thomas "Cakeisalie5" Touhey <thomas@touhey.fr>
# This file is part of the textoutpc project, which is MIT-licensed.
#******************************************************************************
""" Base classes to use with tags in textoutpc, with a manager class.

	For your tag to be used as a textoutpc tag, you have to make it
	inherit one of the `TextoutBlockTag` or `TextoutInlineTag` classes.

	Making separate tag modules is possible through the manager class,
	which allows not to hardcode the tags into the module. """

from functools import partial as _p
from inspect import ismodule as _ismod, isclass as _isclass, \
	getargspec as _getargspec, getfullargspec as _getfullargspec, \
	currentframe as _currentframe, getouterframes as _getouterframes
from importlib import import_module as _importmod

__all__ = ["TextoutTags", "TextoutTag", "TextoutBlockTag", "TextoutInlineTag",
	"TextoutParagraphTag"]

def _getargscount(func):
	try:
		return len(_getfullargspec(func).args)
	except:
		return len(_getargspec(func).args)

# ---
# Main base tag class.
# For more about defining a tag, see `/TAGS.md`.
# ---

class TextoutTag:
	""" The textout tag base class.
		Is initialized with these values:
	
		<name><content><name>
			| name: "<name>" (only special chars such as `)
			| value: None
		[<name>]<content>[/<name>]
			| name: "[<name>]"
			| value: None
		[<name>]<content>[/] (when possible)
			| name: "[<name>]"
			| value: None
		[<name>=<value>]<content>[/<name>]
			| name: "[<name>]"
			| value: "<value>"
		[<name>=<value>]<content>[/] (when possible)
			| name: "[<name>]"
			| value: "<value>" """

	aliases = ()

	def __init__(self, name, value, ot, tweaks):
		""" Initialize the textout tag with the documented members. """

		# Store internal data.

		self.__output_type = ot
		self.__tweaks = tweaks

		self.output_type = ot

		# Call both prepare functions.

		if hasattr(self, 'prepare'):
			try:
				assert _getargscount(self.prepare) == 4
				args = (name, value, ot)
			except:
				args = (name, value)
			self.prepare(*args)
		if hasattr(self, 'prepare_' + ot):
			prep = getattr(self, 'prepare_' + ot)
			try:
				assert len(_getargspec(prep).args) == 4
				args = (name, value, ot)
			except:
				args = (name, value)
			prep(*args)

		# Prepare the preprocessing elements.
		if hasattr(self, 'preprocess'):
			if hasattr(self, 'preprocess_' + ot):
				self.__preprocess0 = self.preprocess
				self.preprocess = self.__preprocess_double
		elif hasattr(self, 'preprocess_' + ot):
			self.preprocess = getattr(self, 'preprocess_' + ot)

		if hasattr(self, 'preprocess'):
			self.__preprocess2 = self.preprocess
			self.preprocess = self.__preprocess_and_prepare
		else:
			self.__after_preprocess()

		if hasattr(self, 'default_' + ot):
			self.default = getattr(self, 'default_' + ot)

	def __preprocess_double(self, content):
		""" Preprocess using the two methods. """

		ct = self.__preprocess0(content)
		if ct != None: content = ct; del ct
		ct = self.__preprocess1(content)
		if ct != None: content = ct; del ct
		return content

	def __preprocess_and_prepare(self, content):
		""" Preprocess and do the things after. """

		ret = self.__preprocess2(content)
		self.__after_preprocess()
		return ret

	def __out(self, name):
		""" Generic function to call two output functions of the same
			type. """

		getattr(self, '__' + name)()
		getattr(self, name + '_' + self.__output_type)()

	def __after_preprocess(self):
		""" After preprocessing, check the begin, content and end that may
			have been set by the preprocessing function. """

		ot = self.__output_type

		for otype in ('begin', 'content', 'end'):
			if hasattr(self, otype):
				if hasattr(self, otype + '_' + ot):
					setattr(self, '__' + otype, getattr(self, otype))
					setattr(self, otype, _p(self.__out, otype))
			elif hasattr(self, otype + '_' + ot):
				setattr(self, otype, getattr(self, otype + '_' + ot))

	def tweak(self, key, default = None):
		try:
			return self.__tweaks[key]
		except KeyError:
			return default

# ---
# Role-specific base tag classes.
# ---

class TextoutBlockTag(TextoutTag):
	pass
class TextoutInlineTag(TextoutTag):
	pass

# ---
# Default tag: paragraph.
# ---

class TextoutParagraphTag(TextoutBlockTag):
	""" Main tag for basic paragraphs. """

	notempty = True

	def begin_html(self):
		return '<p>'

	def end_html(self):
		return '</p>'

# ---
# Tag extractor.
# ---

class TextoutTags:
	""" Tag manager.
		Object responsible for getting the tags. """

	def __init__(self, *modules):
		self._aliases = {}
		for mod in modules:
			self.import_tags(mod)

	def __extract_tags(self, module):
		""" Extract tags from a module. """

		tags = []

		# Obtain the list of properties from the module.

		try:
			ds = module.__all__
		except:
			ds = dir(module)

		# Get the submodules from the module (usually different files in the
		# tags module folder).

		for submodule in (obj for name, obj in ((nm, getattr(module, nm)) \
		for nm in ds) if (name == '__init__' or name[0] != '_') \
		and _ismod(obj)):
			obtained = self.__extract_tags(submodule)
			tags += [tag for tag in obtained \
				if not any(tag is x for x in tags)]
			del obtained

		# Extract the tags from the current module.

		for tag in (obj for name, obj in ((nm, getattr(module, nm)) \
			for nm in ds) if name[0] != '_' and _isclass(obj) \
			and issubclass(obj, TextoutTag)):
			tags.append(tag)

		return tags

	def import_tags(self, module):
		""" Import tags from a dedicated module. """

		if not _ismod(module):
			module = _importmod(module,
				_getouterframes(_currentframe(), 1)[0].name)
		for tag in self.__extract_tags(module):
			for alias in tag.aliases:
				self._aliases[alias] = tag

	def get_tag(self, name, value, output_type = 'html', tweaks = {}):
		""" Initialize a tag. """

		try:
			als = self._aliases[name]
			return als(name, value, output_type, tweaks)
		except:
			return None

# End of file.



#!/usr/bin/env python3
#******************************************************************************
# Copyright (C) 2018 Thomas "Cakeisalie5" Touhey <thomas@touhey.fr>
# This file is part of the textoutpc project, which is MIT-licensed.
#******************************************************************************

from ..tags import TextoutInlineTag as _TextoutInlineTag
import string as _string

__all__ = ["TextoutRotTag"]

class TextoutRotTag(_TextoutInlineTag):
	""" Tag which un-rot13 a content.
		Demonstration tag for content processing.
		Example uses:

		[rot=13]obawbhe[/rot]
		[rot13]Obawbhe[/rot13]
	"""

	aliases = ('[rot]', '[rot13]')
	raw = True

	def prepare(self, name, value):
		if name == "[rot]":
			rot = int(value)
			assert 1 <= rot <= 25
		else:
			rot = int(name[4:-1])

		upr0 = _string.ascii_uppercase
		upr1 = upr0[rot:] + upr0[:rot]
		lwr0 = _string.ascii_lowercase
		lwr1 = lwr0[rot:] + lwr0[:rot]
		self._trans = str.maketrans(upr0 + lwr0, upr1 + lwr1)

	def preprocess(self, content):
		return str.translate(content, self._trans)

# End of file.



#!/usr/bin/env python3
#******************************************************************************
# Copyright (C) 2018 Thomas "Cakeisalie5" Touhey <thomas@touhey.fr>
# This file is part of the textoutpc project, which is MIT-licensed.
#******************************************************************************

from ..tags import TextoutBlockTag as _TextoutBlockTag
from html import escape as _htmlescape
from ..smileys import htmlsmileys as _htmlsmileys

__all__ = ["TextoutSpoilerTag"]

class TextoutSpoilerTag(_TextoutBlockTag):
	""" Hide content at first glance, force people to click to read content.
		These elements can contain 'secret' elements such as solutions, source
		code, or various other things.
		Example uses:
		
		[spoiler]This is hidden![/spoiler]
		[spoiler=Y a quelque chose de caché !|Ah, bah en fait non :)]:E
		And it's multiline, [big]and formatted[/big], as usual :D[/spoiler]
	"""

	aliases = ('[spoiler]',)
	superblock = True
	notempty = True

	def prepare(self, name, value):
		self._closed = "Cliquez pour découvrir"
		self._open = "Cliquez pour recouvrir"

		if value:
			titles = value.split('|')
			if titles[0]:
				self._closed = titles[0]
			if len(titles) >= 2 and titles[1]:
				self._open = titles[1]

	def begin_html(self):
		return '<div class="spoiler">' \
		'<div class="title on" onclick="toggleSpoiler(this.parentNode, ' \
			'\'open\');"><p>{}</p></div>' \
		'<div class="title off" onclick="toggleSpoiler(this.parentNode, ' \
			'\'close\');"><p>{}</p></div>' \
		'<div class="off">'.format(_htmlsmileys(_htmlescape(self._closed)),
			_htmlsmileys(_htmlescape(self._open)))

	def end_html(self):
		return '</div></div>'

# End of file.



#!/usr/bin/env python3
#******************************************************************************
# Copyright (C) 2018 Thomas "Cakeisalie5" Touhey <thomas@touhey.fr>
# This file is part of the textoutpc project, which is MIT-licensed.
#******************************************************************************
""" Autolinking (URL extraction from raw text) in HTML. """

import regex as _re

__all__ = ["htmlurls", "lightscripturls"]

# ---
# Autolinking regex.
# ---

def _sub_html(m):
	sp = m.group('sp')
	url = m.group('url')
	aft = ''

	# Hack for the last comma.
	if url[-1] == ',':
		url, aft = url[:-1], ','

	text = '{}<a href="{}">{}</a>{}' \
		.format(sp, url, url, aft)
	return text

def _sub_lightscript(m):
	sp = m.group('sp')
	url = m.group('url')
	aft = ''

	# Hack for the last comma.
	if url[-1] == ',':
		url, aft = url[:-1], ','

	url = url.replace('<', '%3C')
	url = url.replace('>', '%3E')
	text = '{}<{}>{}'.format(sp, url, aft)
	return text

_reg = _re.compile("""\
	(?P<sp>^|\s|[[:punct:]])
	(?P<url>(https?|ftp):
		(?P<ucore>[^\[\]\(\)\s]* (\[(?&ucore)\]?)* (\((?&ucore)\)?)*)*
	)
""", _re.VERBOSE | _re.M)

# ---
# Main functions.
# ---

def htmlurls(text):
	return _reg.sub(_sub_html, text)

def lightscripturls(text):
	return _reg.sub(_sub_lightscript, text)

# End of file.



#!/usr/bin/env python3
#******************************************************************************
# Copyright (C) 2018 Thomas "Cakeisalie5" Touhey <thomas@touhey.fr>
# This file is part of the textoutpc project, which is MIT-licensed.
#******************************************************************************
""" Unit tests for the `textoutpc` Python module. """

# This file is only there to indicate that the folder is a module.
# It doesn't actually contain code.

# End of file.



#!/usr/bin/env python3
#******************************************************************************
# Copyright (C) 2018 Thomas "Cakeisalie5" Touhey <thomas@touhey.fr>
# This file is part of the textoutpc project, which is MIT-licensed.
#******************************************************************************

from ..tags import TextoutInlineTag as _TextoutInlineTag
from html import escape as _htmlescape

__all__ = ["TextoutLinkTag", "TextoutProfileTag"]

class TextoutLinkTag(_TextoutInlineTag):
	""" The main link tag.
		Example uses:
		
		[url=https://example.org/hi]Go to example.org[/url]!
		[url=/Fr/index.php][/url]
		[url]https://random.org/randomize.php[/url]
	"""

	aliases = ('[url]',)
	raw = True

	def _validate(self):
		for prefix in ('http://', 'https://', 'ftp://', '/', '#'):
			if self._url.startswith(prefix):
				break
		else:
			raise Exception("No allowed prefix!")

	def prepare(self, name, value):
		self._url = None

		# If there is no value, wait until we have a content to
		# decide if we are valid or not.

		if value == None:
			self.preprocess = self._preprocess_if_no_value
			return

		# Otherwise, get the URL and validate.

		self._url = value
		self._validate()
		self.default = self._default_if_value

	def _default_if_value(self):
		return self._url

	def _preprocess_if_no_value(self, content):
		self._url = content
		self._validate()

	def begin_html(self):
		return '<a href="{}">'.format(_htmlescape(self._url))

	def end_html(self):
		return '</a>'

	def begin_lightscript(self):
		return '['

	def end_lightscript(self):
		url = self._url.replace('(', '%28').replace(')', '%29')
		return ']({})'.format(url)

class TextoutProfileTag(TextoutLinkTag):
	""" A special link tag for Planète Casio's profiles.
		Adds the prefix to the content, and sets the value.
		Example uses:
		
		[profil]Cakeisalie5[/profil]
	"""

	aliases = ('[profil]', '[profile]')

	def prepare(self, name, value):
		# Override the TextoutLinkTag's prepare method.

		pass

	def preprocess(self, content):
		# FIXME: check the username content!

		username = content
		self._url = 'https://www.planet-casio.com/Fr/compte/voir_profil.php' \
			'?membre={}'.format(username)
		self._validate()

# End of file.



#!/usr/bin/env python3
#******************************************************************************
# Copyright (C) 2018 Thomas "Cakeisalie5" Touhey <thomas@touhey.fr>
# This file is part of the textoutpc project, which is MIT-licensed.
#******************************************************************************
""" Conversions between color systems. """

from colorsys import hls_to_rgb

__all__ = ["hls_to_rgb", "hwb_to_rgb"]

def hwb_to_rgb(hue, w, b):
	""" Convert HWB to RGB color.
		https://drafts.csswg.org/css-color/#hwb-to-rgb """

	r, g, b = hls_to_rgb(hue, 0.5, 1.0)
	f = lambda x: x * (1 - w - b) + w
	r, g, b = f(r), f(g), f(b)

	return r, g, b

# End of file.



#!/usr/bin/env python3
#******************************************************************************
# Copyright (C) 2018 Thomas "Cakeisalie5" Touhey <thomas@touhey.fr>
# This file is part of the textoutpc project, which is MIT-licensed.
#******************************************************************************

from ..tags import TextoutInlineTag as _TextoutInlineTag
import re as _re

__all__ = ["TextoutLabelTag", "TextoutTargetTag"]

_labelexpr = _re.compile('^[a-z0-9-]{1,16}$', _re.I)

class TextoutLabelTag(_TextoutInlineTag):
	""" The label tag, defines an anchor at a point of the post.
		Example uses:
		
		[label=installation]Installation de tel logiciel... (no ending req.)
		[label=compilation][/label] Compilation de tel logiciel...
	"""

	aliases = ('[label]',)

	def prepare(self, name, value):
		if not _labelexpr.match(value):
			raise Exception
		self._label = value

	def begin_html(self):
		#name = 'label-{}'.format(self._label)
		#if _v42compat:
		#	name += ' ' + self._label
		name = self.tweak("label_prefix", "") + self._label
		return '<a name="{}"></a>'.format(name)

class TextoutTargetTag(_TextoutInlineTag):
	""" The goto tag, links to an anchor defined in the post.
		Example uses:
		
		[target=installation]Check out the installation manual[/target]!
	"""

	aliases = ('[target]',)

	def prepare(self, name, value):
		if not _labelexpr.match(value):
			raise Exception
		self._label = value

	def begin_html(self):
		#name = 'label-' + self._label
		name = self.tweak("label_prefix", "") + self._label
		return '<a href="#{}">'.format(name)

	def end_html(self):
		return '</a>'

# End of file.



#!/usr/bin/env python3
#******************************************************************************
# Copyright (C) 2018 Thomas "Cakeisalie5" Touhey <thomas@touhey.fr>
# This file is part of the textoutpc project, which is MIT-licensed.
#******************************************************************************
""" Smiley conversion.
	Just convert them™.
"""

import regex as _re
from html import escape as _htmlescape

__all__ = ["htmlsmileys"]

# ---
# List of them.
# ---

_Smileys_prefix = "/images/smileys/"

_Smileys = {
	'>:)':			'twisted.gif',
	'>:(':			'evil.gif',
	':)':			'smile.gif',
	';)':			'wink.gif',
	':(':			'sad.gif',
	':D':			'grin.gif',
	':p': 			'hehe.gif',
	'8-)':			'cool2.gif',
	':@':			'mad.gif',
	'0_0':			'eek.gif',
	':E':			'mrgreen.gif',
	':O':			'shocked.gif',
	':s':			'confused2.gif',
	'^^':			'eyebrows.gif',
	":'(":			'cry.gif',
#	':-°':			('whistle.gif', 'height: 15px;'),

	# Name-based smileys.

	':lol:':		'lol.gif',
	':oops:':		'confused2.gif',
	':grr:':		'evil.gif',
	':sry:':		'redface.gif',
	':mmm:':		'rolleyes.gif',
	':waza:':		'waza.gif',
#	':whistle:':	('whistle.gif', 'height: 15px;'),
	':here:':		'pointer.gif',
	':bow:':		'bow.gif',
	':cool:':		'cool.gif',
	':good:':		'welldone.gif',
	':love:':		'love.gif',
	':aie:':		'banghead2.gif',
	':cry:':		'cry.gif',
	':facepalm:':	'facepalm.gif',
	':argh:':		'insults.gif',
	':?:':			'what.gif',
	':!:':			'excl.gif',
	':arrow:':		'here.gif',
	':grin:':		'grin.gif',
}
_Smileys_html = {_htmlescape(a): _Smileys_prefix + b \
	for a, b in _Smileys.items()}

def _Smiley_sub_html(m):
	return m.group(1) + '<img src="' + _Smileys_html[m.group(2)] \
		+ '">' + m.group(3)

_Smiley_html_re = _re.compile('(^|\\s)(' + '|'.join(map(_re.escape,
	_Smileys_html.keys())) + ')(\\s|$)')

# ---
# Functions.
# ---

def htmlsmileys(text):
	""" HTML smileys """

	text = _Smiley_html_re.sub(_Smiley_sub_html, text)
	text = _Smiley_html_re.sub(_Smiley_sub_html, text)
	return text

# End of file.



#!/usr/bin/env python3
#******************************************************************************
# Copyright (C) 2018 Thomas "Cakeisalie5" Touhey <thomas@touhey.fr>
# This file is part of the textoutpc project, which is MIT-licensed.
#******************************************************************************
""" HTML/CSS-like color parsing, mainly for the `[color]` tag.
	Defines the `get_color()` function which returns an rgba value.

	The functions in this module do not aim at being totally compliant with
	the W3C standards, although it is inspired from it.
"""

from .read import get_color

__all__ = ["get_color"]

# End of file.



#!/usr/bin/env python3
#******************************************************************************
# Copyright (C) 2018 Thomas "Cakeisalie5" Touhey <thomas@touhey.fr>
# This file is part of the textoutpc project, which is MIT-licensed.
#******************************************************************************

from ..tags import TextoutBlockTag as _TextoutBlockTag

__all__ = ["TextoutTitleTag"]

class TextoutTitleTag(_TextoutBlockTag):
	""" The title tag.
		Example uses:
		
		[title]Some title[/title]
		[subtitle]Some subtitle[/subtitle]
	"""

	aliases = ('[title]', '[subtitle]')
	raw = True

	def prepare(self, name, value):
		self._level = name[1:-1]

	def begin_html(self):
		return ('<h5>', '<h4>')[self._level == "title"]

	def end_html(self):
		return ('</h5>', '</h4>')[self._level == "title"]

	def begin_lightscript(self):
		return '#' * ((self._level == "subtitle") + 1) + ' '

# End of file.



#!/usr/bin/env python3
#******************************************************************************
# Copyright (C) 2018 Thomas "Cakeisalie5" Touhey <thomas@touhey.fr>
# This file is part of the textoutpc project, which is MIT-licensed.
#******************************************************************************
""" Named colors definitions. Color names are case-insensitive.
	Taken from: https://www.w3schools.com/cssref/css_colors.asp """

__all__ = ["colors"]

colors = {
	# Standard CSS3 named colors, ordered by hex code.

	'black':                '#000000',
	'navy':                 '#000080',
	'darkblue':             '#00008B',
	'mediumblue':           '#0000CD',
	'blue':                 '#0000FF',
	'darkgreen':            '#006400',
	'green':                '#008000',
	'teal':                 '#008080',
	'darkcyan':             '#008B8B',
	'deepskyblue':          '#00BFFF',
	'darkturquoise':        '#00CED1',
	'mediumspringgreen':    '#00FA9A',
	'lime':                 '#00FF00',
	'springgreen':          '#00FF7F',
	'aqua':                 '#00FFFF',
	'cyan':                 '#00FFFF',
	'midnightblue':         '#191970',
	'dodgerblue':           '#1E90FF',
	'lightseagreen':        '#20B2AA',
	'forestgreen':          '#228B22',
	'seagreen':             '#2E8B57',
	'darkslategray':        '#2F4F4F',
	'darkslategrey':        '#2F4F4F',
	'limegreen':            '#32CD32',
	'mediumseagreen':       '#3CB371',
	'turquoise':            '#40E0D0',
	'royalblue':            '#4169E1',
	'steelblue':            '#4682B4',
	'darkslateblue':        '#483D8B',
	'mediumturquoise':      '#48D1CC',
	'indigo':               '#4B0082',
	'darkolivegreen':       '#556B2F',
	'cadetblue':            '#5F9EA0',
	'cornflowerblue':       '#6495ED',
	'rebeccapurple':        '#663399',
	'mediumaquamarine':     '#66CDAA',
	'dimgray':              '#696969',
	'dimgrey':              '#696969',
	'slateblue':            '#6A5ACD',
	'olivedrab':            '#6B8E23',
	'slategray':            '#708090',
	'slategrey':            '#708090',
	'lightslategray':       '#778899',
	'lightslategrey':       '#778899',
	'mediumslateblue':      '#7B68EE',
	'lawngreen':            '#7CFC00',
	'chartreuse':           '#7FFF00',
	'aquamarine':           '#7FFFD4',
	'maroon':               '#800000',
	'purple':               '#800080',
	'olive':                '#808000',
	'gray':                 '#808080',
	'grey':                 '#808080',
	'skyblue':              '#87CEEB',
	'lightskyblue':         '#87CEFA',
	'blueviolet':           '#8A2BE2',
	'darkred':              '#8B0000',
	'darkmagenta':          '#8B008B',
	'saddlebrown':          '#8B4513',
	'darkseagreen':         '#8FBC8F',
	'lightgreen':           '#90EE90',
	'mediumpurple':         '#9370DB',
	'darkviolet':           '#9400D3',
	'palegreen':            '#98FB98',
	'darkorchid':           '#9932CC',
	'yellowgreen':          '#9ACD32',
	'sienna':               '#A0522D',
	'brown':                '#A52A2A',
	'darkgray':             '#A9A9A9',
	'darkgrey':             '#A9A9A9',
	'lightblue':            '#ADD8E6',
	'greenyellow':          '#ADFF2F',
	'paleturquoise':        '#AFEEEE',
	'lightsteelblue':       '#B0C4DE',
	'powderblue':           '#B0E0E6',
	'firebrick':            '#B22222',
	'darkgoldenrod':        '#B8860B',
	'mediumorchid':         '#BA55D3',
	'rosybrown':            '#BC8F8F',
	'darkkhaki':            '#BDB76B',
	'silver':               '#C0C0C0',
	'mediumvioletred':      '#C71585',
	'indianred':            '#CD5C5C',
	'peru':                 '#CD853F',
	'chocolate':            '#D2691E',
	'tan':                  '#D2B48C',
	'lightgray':            '#D3D3D3',
	'lightgrey':            '#D3D3D3',
	'thistle':              '#D8BFD8',
	'orchid':               '#DA70D6',
	'goldenrod':            '#DAA520',
	'palevioletred':        '#DB7093',
	'crimson':              '#DC143C',
	'gainsboro':            '#DCDCDC',
	'plum':                 '#DDA0DD',
	'burlywood':            '#DEB887',
	'lightcyan':            '#E0FFFF',
	'lavender':             '#E6E6FA',
	'darksalmon':           '#E9967A',
	'violet':               '#EE82EE',
	'palegoldenrod':        '#EEE8AA',
	'lightcoral':           '#F08080',
	'khaki':                '#F0E68C',
	'aliceblue':            '#F0F8FF',
	'honeydew':             '#F0FFF0',
	'azure':                '#F0FFFF',
	'sandybrown':           '#F4A460',
	'wheat':                '#F5DEB3',
	'beige':                '#F5F5DC',
	'whitesmoke':           '#F5F5F5',
	'mintcream':            '#F5FFFA',
	'ghostwhite':           '#F8F8FF',
	'salmon':               '#FA8072',
	'antiquewhite':         '#FAEBD7',
	'linen':                '#FAF0E6',
	'lightgoldenrodyellow': '#FAFAD2',
	'oldlace':              '#FDF5E6',
	'red':                  '#FF0000',
	'magenta':              '#FF00FF',
	'fuchsia':              '#FF00FF',
	'deeppink':             '#FF1493',
	'orangered':            '#FF4500',
	'tomato':               '#FF6347',
	'hotpink':              '#FF69B4',
	'coral':                '#FF7F50',
	'darkorange':           '#FF8C00',
	'lightsalmon':          '#FFA07A',
	'orange':               '#FFA500',
	'lightpink':            '#FFB6C1',
	'pink':                 '#FFC0CB',
	'gold':                 '#FFD700',
	'peachpuff':            '#FFDAB9',
	'navajowhite':          '#FFDEAD',
	'moccasin':             '#FFE4B5',
	'bisque':               '#FFE4C4',
	'mistyrose':            '#FFE4E1',
	'blanchedalmond':       '#FFEBCD',
	'papayawhip':           '#FFEFD5',
	'lavenderblush':        '#FFF0F5',
	'seashell':             '#FFF5EE',
	'cornsilk':             '#FFF8DC',
	'lemonchiffon':         '#FFFACD',
	'floralwhite':          '#FFFAF0',
	'snow':                 '#FFFAFA',
	'yellow':               '#FFFF00',
	'lightyellow':          '#FFFFE0',
	'ivory':                '#FFFFF0',
	'white':                '#FFFFFF',

	# Keyword/special named color.

	'transparent': 'rgba(0,0,0,0)',
}

# End of file.



#!/usr/bin/env python3
#******************************************************************************
# Copyright (C) 2018 Thomas "Cakeisalie5" Touhey <thomas@touhey.fr>
# This file is part of the textoutpc project, which is MIT-licensed.
#******************************************************************************
""" Built-in tags for the `textoutpc` module.
	Some of these tags will probably have to move to a separate module
	Planète Casio-specific, but still, here we are.
"""

_names = ["Align", "Code", "Image", "Label", "Link", "Progress",
	"Quote", "Rot", "Show", "Spoiler", "Text", "Title", "Video"]
for name in _names:
	__import__(name, globals(), level=1)

# End of file.



#!/usr/bin/env python3
#******************************************************************************
# Copyright (C) 2018 Thomas "Cakeisalie5" Touhey <thomas@touhey.fr>
# This file is part of the textoutpc project, which is MIT-licensed.
#******************************************************************************

from ..tags import TextoutBlockTag as _TextoutBlockTag
from html import escape as _htmlescape

__all__ = ["TextoutImageTag", "TextoutAdminImageTag"]

class TextoutImageTag(_TextoutBlockTag):
	""" The main tag for displaying an image.
		Example uses:
	
		[img]picture_url[/img]
		[img=center]picture_url[/img]
		[img=12x24]picture_url[/img]
		[img=center|12x24]picture_url[/img]
		[img=x24|right]picture_url[/img]
	"""

	aliases = ('[img]',)
	raw = True

	def prepare(self, name, value):
		self._width = None
		self._height = None
		self._align = None
		self._float = False

		for arg in ("", value)[value != None].split('|'):
			if   not arg:
				pass
			elif arg[0] in '0123456789x':
				self._width = None
				self._height = None

				dim = arg.split('x')
				try:    self._width = int(dim[0])
				except: pass
				try:    self._height = int(dim[1])
				except: pass
			elif arg in ('center', 'left', 'right'):
				self._align = arg
			elif arg in ('float-left', 'float-right'):
				self._align = arg[6:]
				self._float = True
			elif arg in ('float', 'floating'):
				self._float = True

	def preprocess(self, content):
		for prefix in ('http://', 'https://', 'ftp://', '/'):
			if content.startswith(prefix):
				break
		else:
			raise Exception("No allowed prefix!")

		self._url = content

	def content_html(self):
		style = []
		cls = []
		if   self._width:
			style.append('width: {}px'.format(self._width))
		elif self._height:
			style.append('width: auto')
		if   self._height:
			style.append('height: {}px'.format(self._height))
		elif self._width:
			style.append('height: auto')
		if   self._float:
			cls.append('img-float')
		if   self._align:
			cls.append('img-{}'.format(self._align))

		return '<img src="{}"{}{} />'.format(\
			_htmlescape(self._url),
			' class="{}"'.format(' '.join(cls)) if cls else '',
			' style="{}"'.format('; '.join(style)) if style else '')

	def content_lightscript(self):
		url = self._url.replace('[', '%5B').replace(']', '%5D')
		return '[[image:{}]]'.format(url)

class TextoutAdminImageTag(TextoutImageTag):
	""" This tag is special for Planète Casio, as it takes images from
		the `ad`ministration's image folder.
		It just adds this folder's prefix.
		Example uses:
		
		[adimg]some_picture.png[/img]
		[adimg=center]some_picture.png[/img]
		[adimg=12x24]some_picture.png[/img]
		[adimg=center|12x24]some_picture.png[/img]
		[adimg=x24|right]some_picture.png[/img]
	"""

	aliases = ('[adimg]',)

	def preprocess(self, content):
		path = content
		# FIXME: check image URL!
		self._url = 'https://www.planet-casio.com/images/ad/' + path

# End of file.



#!/usr/bin/env python3
#******************************************************************************
# Copyright (C) 2018 Thomas "Cakeisalie5" Touhey <thomas@touhey.fr>
# This file is part of the textoutpc project, which is MIT-licensed.
#******************************************************************************

from ..tags import TextoutBlockTag as _TextoutBlockTag
from html import escape as _htmlescape
from ..smileys import htmlsmileys as _htmlsmileys

__all__ = ["TextoutQuoteTag"]

class TextoutQuoteTag(_TextoutBlockTag):
	""" The main tag for quoting someone.
		Example uses:
		
		[quote]Hey, I said that![/quote]
		[quote=Someone important]I said something important, and it's
		multiline and [b]formatted[/b]!
		[quote=Someone else]Heck, he's even quoting me in his quote![/quote]
		[/quote]
	"""

	aliases = ('[quote]',)
	superblock = True
	notempty = True

	def prepare(self, name, value):
		self._value = value

	def begin_html(self):
		f = '<div class="citation">'
		if self._value:
			f += '<p><b>{} a écrit :</b></p>' \
				.format(_htmlsmileys(_htmlescape(self._value)))
		return f

	def end_html(self):
		return '</div>'

	def begin_lightscript(self):
		text = '<<<'
		if self._value:
			text += ' ' + self._value
		return text + '\n'

	def end_lightscript(self):
		return '<<<\n'

# End of file.



#!/usr/bin/env python3
#******************************************************************************
# Copyright (C) 2018 Thomas "Cakeisalie5" Touhey <thomas@touhey.fr>
# This file is part of the fingerd Python 3.x module, which is MIT-licensed.
#******************************************************************************
""" Setup script for the textoutpc Python package and script. """

from sphinx.setup_command import BuildDoc as _BuildDoc
from setuptools import setup as _setup

# Actually, most of the project's data is read from the `setup.cfg` file.

_setup(cmdclass = {'build_sphinx': _BuildDoc})

# End of file.



#!/usr/bin/env python3
#******************************************************************************
# Copyright (C) 2018 Thomas "Cakeisalie5" Touhey <thomas@touhey.fr>
# This file is part of the textoutpc project, which is MIT-licensed.
#******************************************************************************

from ..tags import TextoutBlockTag as _TextoutBlockTag
from html import escape as _htmlescape

__all__ = ["TextoutShowTag"]

class TextoutShowTag(_TextoutBlockTag):
	""" Tag which shows the HTML code that is produced by textout().
		Example uses:
		
		[show][b]hello world![/show]
	"""

	aliases = ('[show]',)
	notempty = True
	superblock = True
	inlined = True
	generic = False
	raw = False

	def preprocess_html(self, content):
		return _htmlescape(content)

	def begin_html(self):
		return '<span style="font-family: monospace;">'

	def end_html(self):
		return '</span>'

# End of file.



#!/usr/bin/env python3
#******************************************************************************
# Copyright (C) 2018 Thomas "Cakeisalie5" Touhey <thomas@touhey.fr>
# This file is part of the textoutpc project, which is MIT-licensed.
#******************************************************************************
""" Unit tests for the Python version of textout.
	Uses the builtin `unittest` module.
"""

import unittest
import textoutpc

# Define the tests.

__test_cases = {
	# Basic text.
	'': '',
	'lol': '<p>lol</p>',
	'<script>alert(1);</script>': \
		'<p>&lt;script&gt;alert(1);&lt;/script&gt;</p>',

	# Other tests. (?)
	'[a][c][/a]': '<p>[a][c][/a]</p>',
	'[a][a]': '<p>[a][a]</p>',
	"[<>]><[/<>]": "<p>[&lt;&gt;]&gt;&lt;[/&lt;&gt;]</p>",

	# Autolinking.
	'(http://www.example.org/some-[damn-url]-(youknow))': \
		'<p>(<a href="http://www.example.org/some-[damn-url]-(youknow)">' \
		'http://www.example.org/some-[damn-url]-(youknow)</a>)</p>',
	'https://thomas.touhey.fr/, tu vois ?': \
		'<p><a href="https://thomas.touhey.fr/">https://thomas.touhey.fr/' \
		'</a>, tu vois ?</p>',

	# Basic text styling.
	'[u][b][a][i][/b]': "<p><u><b>[a]</b></u></p>",
	'[u][b]a[/]mdr': '<p><u><b>a</b>mdr</u></p>',

	# Blocks, alignment.
	'[left]': '',
	'[left]lol[/]hi': '<div class="align-left"><p>lol</p></div><p>hi</p>',
	'a[justify]b': '<p>a</p><div class="align-justify"><p>b</p></div>',
	'a[i]': '<p>a</p>',
	'a[i][justify]b': '<p>a</p>' \
		'<div class="align-justify"><p><i>b</i></p></div>',
	'a[i]k[center]b': '<p>a<i>k</i></p>' \
		'<div class="align-center"><p><i>b</i></p></div>',
	'a[i]k[center][b]b[justify]c[/center]d[/]wouhou': \
		'<p>a<i>k</i></p>' \
		'<div class="align-center"><p><i><b>b</b></i></p>' \
		'<div class="align-justify"><p><i><b>c</b></i></p></div></div>' \
		'<p><i>d</i>wouhou</p>',

	# Show tag for super preprocessing blocks.
	'[show]lol': '<p><span style="font-family: monospace;">lol</span></p>',
	'[quote][show][justify]hehe': \
		'<div class="citation"><p><span style="font-family: monospace;">' \
		'&lt;div class=&quot;align-justify&quot;&gt;' \
		'&lt;p&gt;hehe&lt;/p&gt;&lt;/div&gt;' \
		'</span></p></div>',

	# Titles.
	'lolk[title]smth': '<p>lolk</p>' '<h4>smth</h4>',
	'[subtitle]<>': '<h5>&lt;&gt;</h5>',

	# Fonts.
	'[arial]test': '<p><span style="font-family: arial">test</span></p>',
	'[font=mono]stereo': \
		'<p><span style="font-family: monospace">stereo</span></p>',
	'[haettenschweiler]': '',
	'[font=hello]yea': '<p>[font=hello]yea</p>',

	# Color.
	'yea[color=blue]dabadee': \
		'<p>yea<span style="color: #0000FF">dabadee</span></p>',
	'[color=#12345F]a': '<p><span style="color: #12345F">a</span></p>',
	'[color=#123]a': '<p><span style="color: #112233">a</span></p>',
	'[color=123]a': '<p><span style="color: #010203">a</span></p>',
	'[color=chucknorris]a': '<p><span style="color: #C00000">a</span></p>',
	'[color=rgb(1, 22,242)]a': '<p><span style="color: #0116F2">a</span></p>',
	'[color= rgb (1,22, 242 , 50.0% )]a': '<p><span style="color: #0116F2; ' \
		'color: rgba(1, 22, 242, 0.5)">a</span></p>',
	'[color=rgba(1,22,242,0.500)]a': '<p><span style="color: #0116F2; ' \
		'color: rgba(1, 22, 242, 0.5)">a</span></p>',
	'[color=hsl(0, 1,50.0%)]r': '<p><span style="color: #FF0000">r</span></p>',
	# TODO: hls, hwb

	# Links.
	'[url]': '<p>[url]</p>',
	'[url=https://thomas.touhey.fr/]mon profil est le meilleur[/url]':
		'<p><a href="https://thomas.touhey.fr/">mon profil est le meilleur' \
		'</a></p>',
	'[url=https://thomas.touhey.fr/]': \
		'<p><a href="https://thomas.touhey.fr/">https://thomas.touhey.fr/' \
		'</a></p>',
	'[url=http://hey.org/lol[]>"a]': '<p><a href="http://hey.org/lol[]&gt;' \
		'&quot;a">' \
		'http://hey.org/lol[]&gt;&quot;a</a></p>',
	'[url]javascript:alert(1)[/url]': '<p>[url]javascript:alert(1)[/url]</p>',
	'[url]<script>alert(1);</script>[/url]': \
		'<p>[url]&lt;script&gt;alert(1);&lt;/script&gt;[/url]</p>',

	'[profil]cake[/profil]': \
		'<p><a href="https://www.planet-casio.com/Fr/compte/voir_profil.php' \
		'?membre=cake">cake</a></p>',
	'[profile]ekac': \
		'<p><a href="https://www.planet-casio.com/Fr/compte/voir_profil.php' \
		'?membre=ekac">ekac</a></p>',

	# Quotes.
	'[quote]': '',
	'[quote]a': \
		'<div class="citation"><p>a</p></div>',
	'[quote=Test 1 :)]lel[/quote]': \
		'<div class="citation"><p><b>Test 1 ' \
		'<img src="/images/smileys/smile.gif"> a écrit :</b></p><p>' \
		'lel</p></div>',

	# Spoilers.
	'[spoiler]': '',
	'[spoiler=Hello|world> :D]Close this, quick![/spoiler]': \
		'<div class="spoiler"><div class="title on" ' \
		'onclick="toggleSpoiler(this.parentNode, ' "'open'" ');"><p>Hello' \
		'</p></div><div class="title off" ' \
		'onclick="toggleSpoiler(this.parentNode, ' "'close'" ');"><p>world' \
		'&gt; <img src="/images/smileys/grin.gif"></p></div>' \
		'<div class="off"><p>Close this, quick!</p></div></div>',

	# Code.
	'[code]': '',
	"`[code]`": '<p><span style="font-family: monospace;">[code]</span></p>',

	'[inlinecode]': '',
	"[inlinecode]`[/inlinecode]": \
		'<p><span style="font-family: monospace;">`</span></p>',

	"[b]a[noeval]b[/b]c[/noeval]d": "<p><b>ab[/b]cd</b></p>",
	"a[noeval]b[noeval]c[/noeval]d[/noeval]e": "<p>ab[noeval]c[/noeval]de</p>",
	"[noeval]``[/noeval]": "<p>``</p>",
	'[noeval]<>[/noeval]': '<p>&lt;&gt;</p>',

	# Pictures.
	'[img]': '<p>[img]</p>',
	'[img]"incroyable<>"[/img]': \
		'<p>[img]&quot;incroyable&lt;&gt;&quot;[/img]</p>',

	# Videos.
	'[video]"><script>alert(1)</script>[/video]': \
		'<p>[video]&quot;&gt;&lt;script&gt;alert(1)&lt;/script&gt;' \
		'[/video]</p>',
	'[video]<script>alert(document.cookie)</script>[/video]': \
		'<p>[video]&lt;script&gt;alert(document.cookie)&lt;/script&gt;' \
		'[/video]</p>',
	'[video]https://www.youtube.com/watch?v=6odDOOyUawY[/video]': \
		'<div class="video-wrapper video-medium"><iframe ' \
		'src="https://www.youtube.com/embed/6odDOOyUawY" ' \
		'frameborder="0" allowfullscreen></iframe></div>',
	'[video]https://www.youtube.com/watch?v=<script>alert(1)</script>': \
		'<p><a href="https://www.youtube.com/watch?v=&lt;script&gt;alert(1)' \
		'&lt;/script&gt;">' \
		'https://www.youtube.com/watch?v=&lt;script&gt;alert(1)' \
		'&lt;/script&gt;</a></p>',

	# Progress bars.
	'[progress=lol]mdr[/progress]': '<p>[progress=lol]mdr[/progress]</p>',

	# Text rotation obfuscation.
	'[rot13]obawbhe[/rot13]': '<p>bonjour</p>',
}

# Define the tests wrapper, and define the classes.

_cnt = 0
_len = len(str(len(__test_cases)))
_templ = """\
  def test_html{n:0>{l}}(self):
    self.assertEqual(textoutpc.tohtml({i}), {r}, \\
    	"for the following text: " + {i})
"""

def _wrap_test(inp, res):
	global _cnt

	_cnt += 1
	return _templ.format(n = _cnt, l = _len, i = repr(inp), r = repr(res))

exec("class TextoutHTMLTest(unittest.TestCase):\n  maxDiff = None\n" + \
	'\n'.join(map(lambda args: _wrap_test(*args), __test_cases.items())),
	globals())

# If run as main script, run the test function.

if __name__ == '__main__':
	unittest.main()

# End of file.



#!/usr/bin/env python3
#******************************************************************************
# Copyright (C) 2018 Thomas "Cakeisalie5" Touhey <thomas@touhey.fr>
# This file is part of the textoutpc project, which is MIT-licensed.
#******************************************************************************
""" Main translation function.
	See the `Translator` class documentation for more information.
"""

import string as _string
from copy import deepcopy as _deepcopy
from html import escape as _htmlescape
from importlib import import_module as _importmod

from .tags import TextoutBlockTag as _TextoutBlockTag, \
    TextoutParagraphTag as _TextoutParagraphTag, TextoutTags as _Tags
from .stream import TextoutStream as _TextoutStream
from .smileys import htmlsmileys as _htmlsmileys
from .urls import htmlurls as _htmlurls

__all__ = ["Translator"]

# Builtin tags.

_builtin_tags = _Tags(_importmod('..builtin_tags', __name__))

# ---
# Tweaks interface.
# ---

class _TweaksDictionary:
	""" Tweaks dictionary. Read-only, and makes sure to match equivalent
		tweak keyword, e.g. `label_prefix`, `LABELPREFIX` and
		`__LaBeL___PRE_FIX__`. """

	def __normalize(self, name):
		return ''.join(c for c in name if c in _string.ascii_letters).lower()

	def __init__(self, base):
		self.__elts = {}

		for kw in base:
			self.__elts[self.__normalize(kw)] = base[kw]

	def __getitem__(self, key):
		return self.__elts[key]

# ---
# Tag data utility.
# ---

class _TagData:
	BLOCK = 1
	INLINE = 2

	def __init__(self, tag, name, full):
		""" Tag data initialization.
			Here, we prepare all of the attributes from the tag's
			after-preparation attributes. """

		# `name` is the name through which the tag has been called.
		# `full` is the full tag beginning mark.

		self.name = name
		self.type = self.BLOCK if isinstance(tag, _TextoutBlockTag) \
			else self.INLINE
		self.full = full

		# Tag beginning displaying.
		# `notempty` is the moment when (and if) to start displaying the
		# tag's code and content.
		# `started` is whether the tag's beginning has been processed,
		# i.e. if the content is no longer processed.

		self.notempty = bool(tag.notempty) if hasattr(tag, 'notempty') \
			else False
		self.started = False

		# `base` is the actual tag object returned by `get_tag()`.

		self.base = tag

		# Flags and properties calculated from the tag's attributes, using the
		# rules given in `TAGS.md`.
		# `ign` is whether the content should be read while the tag is opened.
		# `generic` is whether the tag can be terminated by the generic
		# tag ending mark [/].
		# `raw` is whether the tag's content should be read as raw.
		# `super` is whether the tag is a superblock or not.
		# `inlined` is whether the next block on the same level is turned into
		# a superblock or not.

		self.ign = not hasattr(tag, 'preprocess') and hasattr(tag, 'content')

		self.generic = False if name == None else bool(tag.generic) \
			if hasattr(tag, 'generic') else True

		self.raw = bool(tag.raw) if hasattr(tag, 'raw') \
			else hasattr(tag, 'preprocess')

		self.super = True if hasattr(tag, 'preprocess') else \
			bool(tag.superblock) if hasattr(tag, 'superblock') \
			else False

		self.inlined = bool(tag.inlined) if self.super \
			and hasattr(tag, 'inlined') and bool(tag.inlined) else False

		# Content processing utilities.
		# `last` is the content of the tag. A boolean indicates that we
		# only want to know if the content is empty or not, and a string
		# means we want to get the full content to re-use it later.
		# In order not to manage a third case, even if the tag doesn't care
		# if its content is empty or not, this property should be set to
		# `False`.

		self.last = "" if hasattr(tag, 'preprocess') else False

		# Reset the tag.

		self.reset()

	def reset(self):
		""" Reset the tag, generally because it has been closed. """

		self.tag = _deepcopy(self.base)
		self.started = False
		if isinstance(self.last, bool):
			self.last = False
		else:
			self.last = ""

	def __repr__(self):
		return '<TagData tag=' + repr(self.tag) + '>'

# ---
# Translator main class.
# ---

class Translator:
	""" One-time usage class for translating.
		Use it this way: `Translator(my_inp, my_outp).process()`.

		You can even chain calls as the `process()` method returns
		the output stream object. """

	def __init__(self, inp, outp, output_type = 'html', \
		tweaks = {}, tags = _builtin_tags):
		""" Initializer. """

		if not output_type in ('html', 'lightscript'):
			raise Exception("Invalid output type")
		self.output_type = output_type

		self.tweaks = _TweaksDictionary(tweaks)
		self.tags = tags

		self.inp = inp
		self.outp = outp

		# `queue` is the queue of tag containers, with the actual tag
		# objects, calculated tag properties, variables for content processing,
		# and other stuff.
		# `cign` is the number of tags requiring the content to be ignored.

		self.queue = []
		self.cign = 0

		# Text group management.
		# In the following example text:
		#
		#	some [incredible] text [align=center] you know
		#
		# There are two input groups, what's before and what's after the
		# valid `[align=center]` tag. We want to flush the text in two steps
		# only, in order to detect things such as URLs and smileys.
		#
		# The text group also manages the invalid tags, to manage URLs with
		# brackets in it, e.g. https://example.org/[some-incredible-thing]-yea

		self.text_group = ""

		# `raw_mode` is whether the no evaluating mode is on or not.
		# `raw_deg` is the number of times the raw tag has to be closed
		# to exit.

		self.raw_mode = False
		self.raw_deg = 0

	# ---
	# Text outputting utilities.
	# ---

	def process_text(self, text):
		""" Process text groups for naked URLs and stuff. """

		# In all cases, we want to escape for HTML things, so that the
		# user doesn't insert raw HTML tags (which would be a security flaw!).

		if self.output_type == 'html':
			text = _htmlescape(text)

		# For non-raw HTML, we want to add smiley and URLs conversion,
		# because it's nicer!

		if not self.raw_mode and self.output_type == 'html':
			text = _htmlsmileys(_htmlurls(text))

		return text

	def put_text(self, text):
		""" Output some text. """

		# If we want to ignore the content (because it is not used
		# nor output), let the text fall into the void.

		if self.cign > 0:
			return

		# Add to the text group, which will be processed when `flush_text()`
		# is used.

		self.text_group += text

	def flush_text(self, superblocks_only = False,
		next_block_is_super = False):
		""" Flush the text that has been output. """

		# First of all, check if the text group is empty or if we want to
		# ignore it.

		if not self.text_group or self.cign > 0:
			return

		# Pop the text group and put the code, with the process function in
		# case it is given to a non-raw processing tag or given to the
		# output.

		text = self.text_group
		self.text_group = ""

		self.add_text(text, process_func = lambda x: self.process_text(x),
			superblocks_only = superblocks_only,
			next_block_is_super = next_block_is_super)

	# ---
	# Code outputting utilities.
	# ---

	def add_text(self, text, process_func = lambda x: x, start_tags = True,
		superblocks_only = False, next_block_is_super = False):
		""" Add text to the higher blocks if available. """

		# The last queue is composed of booleans (does the group contain
		# something or not) and texts for content processing.
		# We want to set all of the booleans to True until the first text
		# group, to which we want to add the current text.
		# If there is no content preprocessing and we have to output it,
		# we want to start the tags first: `dat == None` will be our signal!

		blockfound = False
		for dat in self.queue:
			# Check if it is a tag we want to contribute to.

			if dat.type == dat.BLOCK:
				if dat.super or next_block_is_super:
					blockfound = True
					next_block_is_super = dat.inlined
				elif not superblocks_only and not blockfound:
					blockfound = True
					next_block_is_super = dat.inlined
				else:
					continue

			# Contribute to it, either by or-ing the content if it is
			# a boolean (but anything or True == True), or by contributing
			# to the buffer otherwise.

			if isinstance(dat.last, bool):
				dat.last = True
				continue

			# Start the tags if we're about to give this content to
			# preprocessing.

			if start_tags:
				self.start_tags()

			# Add the content to the preprocess buffer.

			if not dat.raw:
				text = process_func(text)
			dat.last += text

			break
		else:
			# No `break` has been encountered, which means the content has
			# not been added to any preprocessing tag. Please process it!

			if start_tags:
				self.start_tags()
			self.outp.write(process_func(text))

			return False

		# The content has been given for preprocessing.

		return True

	def put_debug(self, message):
		""" Put a debug message directly into the output. """

		self.outp.write(message)

	def put_code(self, code, start_tags = True, flush_text = True,
		superblocks_only = True, next_block_is_super = False):
		""" Put some code. """

		# We don't want to mix text and code, so we'll flush to be sure that
		# the order doesn't get mixed up.

		if flush_text:
			self.flush_text()

		# First of all, check if the text is empty or if we want to ignore it.

		if not code or self.cign > 0:
			return

		# Add the code.

		self.add_text(code, start_tags = start_tags,
			superblocks_only = superblocks_only,
			next_block_is_super = next_block_is_super)

	def put_newline(self):
		""" Put a newline. """

		# The newline depends on the output type and the context, of course.

		if self.output_type == 'html' and not self.raw_mode:
			newline = '<br />\n'
		else:
			newline = '\n'

		# Then put this as one puts code.

		self.put_code(newline)

	# ---
	# Tag queue management.
	# ---

	def push_tag(self, dat):
		""" Push a tag onto the tag stack. """

		# If the tag does not process its content but replaces the content,
		# that means the content is ignored.

		if dat.ign:
			self.cign += 1

		# If we're about to put a tag or anything, empty the text block
		# here.

		self.flush_text()

		# Insert the tag into the queue.

		self.queue.insert(0, dat)

		# Start the tag (and parent tags) if required.

		self.start_tags()

		# Don't forget to add the tag to the queue, and to enable raw
		# mode if the tag expects a raw content (e.g. `[code]`).

		if dat.raw:
			self.raw_mode = True
			self.raw_deg = 0

	def pop_tag(self, end = ""):
		""" Pop a tag from the tag stack.
			`end` represents the full version of the ending tag marker,
			for displaying if the tag is invalid. """

		# Even if we had no beginning, no content and no end, what is
		# here has to be distinguished from what was right before!
		# So we need to flush the text group for this.
		# (this will probably be useless for tags with preprocessing enabled,
		# but that's okay, flushing doesn't modify the content processing
		# queue)

		self.flush_text()

		# Pop the tag out of the queue.

		dat = self.queue.pop(0)
		tag = dat.tag

		pcattrs = {'superblocks_only': dat.type == dat.BLOCK,
			'next_block_is_super': dat.inlined}

		# If preprocessing has been enabled, we ought to process the content,
		# check if the tag is valid, and do everything we would have done
		# while pushing the tag if it didn't do content processing.

		if hasattr(tag, 'preprocess'):
			# Take out the content of the content preprocessing queue.
			# If there is no content and the tag proposes a default content,
			# let's use it instead.

			content = dat.last
			if not content and hasattr(tag, 'default'):
				try:
					content = tag.default()
				except:
					# The tag is not supposed to have an empty content,
					# so we ought to put it as an invalid tag an go on.

					self.put_text(dat.full)
					self.put_text(end)
					return

			# Send the content to the tag while checking its validity (by
			# checking if the `preprocess()` method returns an exception).

			try:
				ct = tag.preprocess(content)
			except:
				# The tag is invalid in the end, so we ought to send the
				# raw things to the text group and forget about the tag.

				self.put_text(dat.full)
				self.put_text(content)
				self.put_text(end)
				return

			# If we're here, congrats, the tag is valid! Now, if the
			# `preprocess()` method returned something different, we
			# want to use it instead.

			if ct != None:
				content = ct

			# Output the beginning and the content. If there was no content,
			# just put the content that we got earlier.

			if hasattr(tag, 'begin'):
				self.put_code(tag.begin(), **pcattrs)
			dat.started = True

			if hasattr(tag, 'content'):
				self.put_code(tag.content(), **pcattrs)
			elif dat.raw:
				# XXX: I'm unsure about this. Shall raw tags return code
				# or text? The text will only be escaped as raw mode is
				# still enabled at this point.

				self.put_text(content)
			else:
				self.put_code(content, **pcattrs)
		elif hasattr(tag, 'content'):
			# Tag replaces content without preprocessing, which means
			# the content has been ignored and the tag only puts the
			# things.

			self.cign -= 1
			self.put_code(tag.content(), **pcattrs)
		elif hasattr(tag, 'default'):
			# Tag defines a default content if there might be none,
			# without text preprocessing. If there is no content, print it.
			# Notice that the default content method can also raise
			# an exception if the tag in its current configuration should
			# not have an empty content.

			if not dat.started:
				if hasattr(dat.tag, 'begin'):
					self.put_code(dat.tag.begin(), **pcattrs)
				dat.started = True

			if not dat.last:
				try:
					self.put_text(tag.default())
				except:
					# The tag is not supposed to have empty content!
					# Let's put the raw things again as when there is
					# content processing.

					self.put_text(dat.full)
					self.put_text(end)
					return

		# Don't forget to end the tag!

		if not dat.started:
			pass
		else:
			if dat.type == dat.BLOCK:
				self.close_inline_tags()
			if hasattr(tag, 'end'):
				self.put_code(tag.end(), start_tags = False, **pcattrs)

		# Disable raw mode if it was a raw tag (which means that it enabled it,
		# as tags into raw tags cannot be processed).

		if dat.raw:
			self.raw_mode = False

	# ---
	# Automatically start and end tags.
	# ---

	def start_tags(self):
		""" Start the tags that haven't been started yet.
			If a block has been newly opened, we ought to close the block at
			the same level as them before opening it.
			This is usually called when content is output, for tags that
			aren't empty. """

		# First, get the references to the blocks to end, the blocks to
		# start, and all of the inline tags.

		superblocks = []
		block_to_start = None
		block_to_end = None
		inlines = []

		next_block_is_super = False
		for dat in self.queue:
			# Check if the tag hasn't already been started or doesn't call
			# for content processing.

			if type(dat.last) != bool:
				break

			# Then put the tag in the appropriate queue.

			if dat.type == dat.BLOCK:
				if   dat.super or next_block_is_super:
					superblocks.insert(0, dat)
					next_block_is_super = dat.inlined
				elif dat.started:
					block_to_end = dat
					next_block_is_super = dat.inlined
				elif block_to_end == None and block_to_start == None:
					block_to_start = dat
					next_block_is_super = dat.inlined
			else:
				inlines.insert(0, dat)

		# If there is no new block to start, there's no need to end the
		# current block.

		if not block_to_start:
			block_to_end = None

		# Put the tag ends for the blocks to end.
		# If there are some, we ought to close the inline tags first.

		if block_to_end != None:
			for dat in inlines[::-1] + [block_to_end]:
				if not dat.started:
					continue

				if hasattr(dat.tag, 'end'):
					self.put_code(dat.tag.end(), start_tags = False)
				dat.started = False
				dat.reset()

		# Then, put the tag beginnings.

		to_begin = superblocks \
			+ ([block_to_start] if block_to_start else []) \
			+ inlines

		for dat in to_begin:
			if dat.started:
				continue
			if dat.notempty and not dat.last:
				break

			if hasattr(dat.tag, 'begin'):
				self.put_code(dat.tag.begin(), start_tags = False,
					flush_text = False)
			dat.started = True

	def close_inline_tags(self):
		""" We're about to close a block, so we want to close any inline tags
			that could have been taken within it. """

		for dat in self.queue:
			# Check that the tag hasn't already been closed.

			if dat.type != dat.INLINE or not dat.started:
				continue

			if hasattr(dat.tag, 'end'):
				self.put_code(dat.tag.end(), start_tags = False)
			dat.started = False
			dat.reset()

	# ---
	# Main function.
	# ---

	def process(self):
		""" Main function of the textout translator. """

		# By default, everything is in a paragraph.
		# Other blocks will supplant this by being further in the queue.

		self.push_tag(_TagData(_TextoutParagraphTag(None, None,
			self.output_type, self.tweaks), None, ''))

		# We want to get our elements out of the element stream (Lephe
		# told me that the `TextoutStream` class was actually a lexer,
		# but as I don't know the theory behind this...).

		for element in _TextoutStream(self.inp):
			# If it is a string or a newline, let's just put it.
			# Otherwise, the element is some tag data or at least something
			# that requires some special processing.

			if isinstance(element, str):
				self.put_text(element)
				continue

			tagdata = element
			if tagdata.type == tagdata.NEWLINE:
				self.put_newline()
				continue

			# XXX: As we don't manage paragraphs for now, end of lines and
			# paragraphs separator are just output for now.

			if not tagdata.type in (tagdata.BEGIN, tagdata.END, \
				tagdata.SPECIAL):
				self.put_text(tagdata.full)
				continue

			# Check if it is a tag end (we do not know for special tags,
			# as they usually are one-character long).

			if tagdata.type in (tagdata.END, tagdata.SPECIAL):
				# If raw mode is activated, that means that the queue is not
				# empty and that the top tag of the queue is the tag that
				# initiated raw mode. We're just going to check that the name
				# corresponds, and that the tag has not be opened into
				# itself (see the description of `raw_deg` in the
				# initializer).

				if self.raw_mode:
					if tagdata.name != self.queue[0].name \
					and not (tagdata.name == "[]" and self.queue[0].generic):
						self.put_text(tagdata.full)
						continue
					if self.raw_deg > 0:
						self.put_text(tagdata.full)
						self.raw_deg -= 1
						continue

				# Check to which opened tag the ending tag corresponds.

				pos = -1
				if tagdata.name == "[]":
					# Generic closing tag [/] management.
					# `pos` is set to 0 here.
					for qpos, qdat in enumerate(self.queue):
						if qdat.name != None:
							pos = qpos
							break
				else:
					# Get the position corresponding to the tag.
					for qpos, qdat in enumerate(self.queue):
						if tagdata.name == qdat.name:
							pos = qpos
							break

				# Then react to `pos`.
				# If `pos` is 0 or above, an opening tag has been found.
				# We ought to autoclose opened stuff which are not
				# terminated explicitely, and close the tag closed explicitely.

				if pos >= 0:
					while pos > 0:
						self.pop_tag()
						pos -= 1
					self.pop_tag(tagdata.full)
					continue

				if tagdata.type == tagdata.END:
					self.put_text(tagdata.full)
					continue

				# If we are here, the tag is a special tag which hasn't been
				# identified to be an ending tag. We don't want to stop because
				# that means it is a beginning tag.

			# From here, we know the tag is not a beginning tag.
			# In raw mode, always display the tag, but if the tag corresponds
			# to the raw tag opened, augment the number of tags required to
			# close the raw tag.

			if self.raw_mode:
				if tagdata.name == self.queue[0].name:
					self.raw_deg += 1

				self.put_text(tagdata.full)
				continue

			# Get the initialized tag with the name and value.
			# If the tag is unknown, output the full thing and just go on.

			tag = self.tags.get_tag(tagdata.name, tagdata.value,
				self.output_type, self.tweaks)
			if not tag:
				self.put_text(tagdata.full)
				continue

			# And don't forget to push the tag.

			dat = _TagData(tag, tagdata.name, tagdata.full)
			self.push_tag(dat)

			# Push a paragraph tag if the block is a superblock.

			if dat.type == dat.BLOCK and dat.super and not dat.raw \
				and not dat.inlined:
				self.push_tag(_TagData(_TextoutParagraphTag(None, None,
					self.output_type, self.tweaks), None, ''))

		# End of file, it seems! Let's close the tags and just resume our
		# lives from there.

		while self.queue:
			self.pop_tag()

		# And don't forget to return the output for the user to chain
		# stuff easily ;)

		return self.outp

# End of file.



#!/usr/bin/env python3
#******************************************************************************
# Copyright (C) 2018 Thomas "Cakeisalie5" Touhey <thomas@touhey.fr>
# This file is part of the textoutpc project, which is MIT-licensed.
#******************************************************************************
""" Layer on top of the character stream.

	See the `TextoutStream` class description for more information.
"""

import io as _io
import regex as _re

__all__ = ["TextoutStream", "TextoutUnit"]

# ---
# Class definitions.
# ---

class TextoutUnit:
	""" Raw textout stream unit. """

	BEGIN   = 1
	END     = 2
	SPECIAL = 3
	NEWLINE = 4
	PARSEP  = 5

	def __init__(self, *args):
		if len(args) > 1:
			self.type, self.name, self.value, *_ = args + (None,)
			return

		result = args[0]
		gr = result.groupdict()

		self.name = None
		self.value = None

		if   gr['sname'] == '\n':
			self.type = self.NEWLINE
		elif gr['parsep'] != None:
			self.type = self.PARSEP
		elif gr['bname'] != None:
			self.type = self.BEGIN
			self.name = gr['bname']
			self.value = gr['value']

			self.full = "[{}{}]".format(self.name,
				"=" + self.value if self.value != None else "")
		elif gr['ename'] != None:
			self.type = self.END
			self.name = gr['ename']
			self.full = "[/" + self.name + "]"
		else:
			self.type = self.SPECIAL
			self.name = gr['sname']

			self.full = self.name

		if self.name != None:
			self.name = self.name.lower()
			if self.type != self.SPECIAL:
				self.name = "[{}]".format(self.name)

	def __repr__(self):
		typetab = {self.BEGIN: "begin", self.END: "end",
			self.SPECIAL: "special", self.NEWLINE: "newline"}
		return '<TextoutUnit type={}{}{}>'.format(\
			typetab[self.type],
			' name=' + self.name if self.name != None else "",
			' value="{}"'.format(self.value) if self.value != None else "")

	def __equ__(self, other):
		if not isinstance(other, TextoutUnit):
			return False
		if self.type == other.type \
		and (self.type == self.NEWLINE or self.name == other.name) \
		and (self.type != self.BEGIN or self.value == other.value):
			return False
		return True

class TextoutStream:
	""" Textout stream, for easier stream processing.

		The idea behind this stream is that it will provide more suitable
		(therefore easier to process) data for the applications above,
		with raw text and tags. """

	# A tag can basically be one of the following things:
	# - a starting tag, looking like [<name>] or [<name>=<attribute>]
	# - an ending tag, looking like [/<name>]
	# - a special tag (starting or ending), usually one-char (the only
	#   one currently available is the ` tag).
	#
	# A tag name is 32 chars at most (at least 1 char).
	# A closing tag can have no name, which means that it will close the
	# last opened tag automatically.
	# A tag attribute is 256 chars at most.
	#
	# FIXME: check the sizes? it seems that it stopped working…

	_Tag = _re.compile("""\
		\[\s?(?P<bname>
			(?P<bname_e>[^\/\[\]\=][^\[\]\=]* (\[(?&bname_e)\]?)*)*
		)
		(\s?=\s?(?P<value>
			(?P<value_e>[^\[\]]* (\[(?&value_e)\]?)*)*
		))?\s?\]
	|
		\[[\\/]\s?(?P<ename>
			(?P<ename_e>[^\/\[\]\=][^\[\]\=]* (\[(?&ename_e)\]?)*)*
		)\s?\]
	|
		(?P<parsep>[\n]{2,})
	|
		(?P<sname>`|[\n])
	""", _re.VERBOSE | _re.M)

	# Keep this buffer size above the maximum size of a tag (387)
	# for this class to work alright. Anything above 512 should work great.

	BUFFER_SIZE = 1024

	def __init__(self, stream):
		# If the 'stream' is a string, we want to use standard stream
		# functions, so we're gonna enforce them using the `StringIO` class.

		if isinstance(stream, str):
			stream = _io.StringIO(stream)

		# Buffer management.

		self.stream = stream
		self.buf = ""

		# Management of the last tag match.

		self.result = None
		self.last = None

		# Error position.

		self.pos = 0
		self.line = 0
		self.col = 0

	def __iter__(self):
		# This class is (obviously) iterable.
		# We want to use this class as the iterator as well.

		return self

	def __next__(self):
		# If we have a result, process it.

		if self.result:
			data, self.result = TextoutUnit(self.result), None
			self.last = data
			return data

		# Make sure to have enough data to read.

		self.buf += self.stream.read(self.BUFFER_SIZE - len(self.buf))
		if not self.buf:
			self.last = None
			raise StopIteration

		# Check that we have a result.

		result = self._Tag.search(self.buf, partial = True)
		if not result:
			text = self.buf
			self.buf = ''
			self.last = text
			return text

		# If there is some text, return it.
		# Eventually store the result so we can process it later.

		if result.start() > 0:
			ret = self.buf[:result.start()]
			self.buf = self.buf[result.end():]
			if not result.partial:
				self.result = result
			self.last = ret
			return ret

		# Process the result now!

		self.buf = self.buf[result.end():]
		data = TextoutUnit(result)
		self.last = data
		return data

# End of file.



#!/usr/bin/env python3
#******************************************************************************
# Copyright (C) 2018 Thomas "Cakeisalie5" Touhey <thomas@touhey.fr>
# This file is part of the textoutpc project, which is MIT-licensed.
#******************************************************************************

from ..tags import TextoutBlockTag as _TextoutBlockTag, \
    TextoutInlineTag as _TextoutInlineTag

__all__ = ["TextoutCodeTag", "TextoutInlineCodeTag", "TextoutNoEvalTag"]

class TextoutCodeTag(_TextoutBlockTag):
	""" The basic code tag, for displaying code.
		Example uses:
		
		[code]int main()
		{
			printf("hello, world");
		}[/code] """

	aliases = ('[code]',)
	generic = False
	raw = True
	notempty = True

	def begin_html(self):
		return '<div class="code">'

	def end_html(self):
		return '</div>'

	def begin_lightscript(self):
		return '```\n'

	def end_lightscript(self):
		return '```\n'

class TextoutInlineCodeTag(_TextoutInlineTag):
	""" Inline code tag, doesn't display a box, simply doesn't evaluate
		the content and uses monospace font.
		Example uses:
		
		`some inline code`
		[inlinecode][b]The tags will be shown verbatim.[/b][/inlinecode]
		[inlinecode][inlinecode][i]This also[/inlinecode] works![/inlinecode]
	"""

	aliases = ('`', '[inlinecode]')
	generic = False
	raw = True

	def begin_html(self):
		return '<span style="font-family: monospace;">'

	def end_html(self):
		return '</span>'

	def begin_lightscript(self):
		return '`'

	def end_lightscript(self):
		return '`'

class TextoutNoEvalTag(_TextoutInlineTag):
	""" Inline code tag, simply doesn't evaluate the content.
		Example uses:
		
		[noeval][b]wow, and no need for monospace![/b][/noeval]
	"""

	aliases = ('[noeval]',)
	generic = False
	raw = True

# End of file.



#!/usr/bin/env python3
#******************************************************************************
# Copyright (C) 2018 Thomas "Cakeisalie5" Touhey <thomas@touhey.fr>
# This file is part of the textoutpc project, which is MIT-licensed.
#******************************************************************************

from ..tags import TextoutInlineTag as _TextoutInlineTag
from ..color import get_color

__all__ = ["TextoutTextTag"]

# ---
# Data.
# ---

_big_size = 2.00
_sml_size = 0.75

_fonts = {
	"arial": "Arial",
	"comic": "Comic MS",
	"tahoma": "Tahoma",
	"courier": "Courier",
	"haettenschweiler": "Haettenschweiler",
	"mono": "monospace",
	"monospace": "monospace"
}

# ---
# Tag definition.
# ---

class TextoutTextTag(_TextoutInlineTag):
	""" Main tag for setting text formatting.
		Example uses:
		
		[b]Bold text.[/b]
		[i]Italic text.[/i]
		[u]Underlined text.[/u]
		[strike]Striked text.[/strike]
		[striked]Text strikes again.[/striked]
		[font=arial]Arial text.[/font]
		[arial]Arial text again.[/arial]
		[blue]This will be in blue[/blue]
		[color=blue]This as well[/color]
		[color=rgb(255, 255, 255, 0.4)]BLACKNESS[/color]
		[color=hsl(0, 100%, 0.5)]This will be red.[/color]
	
		Also supports a hack used on Planète Casio for a while, which
		is a CSS injection, e.g.:
	
		[color=brown; size: 16pt]Hello world![/color]
	"""

	aliases = ('[b]', '[i]', '[u]', '[s]', '[strike]',
		'[monospace]', '[mono]', '[font]', '[color]',
		'[size]', '[big]', '[small]',
		'[arial]', '[comic]', '[tahoma]', '[courier]',
		'[haettenschweiler]', '[red]', '[green]', '[blue]',
		'[yellow]', '[maroon]', '[purple]', '[gray]',
		'[grey]', '[brown]')
	notempty = True

	def prepare(self, name, value):
		self._bold = False
		self._italic = False
		self._underline = False
		self._strike = False
		self._font = None
		self._color = None
		self._size = None

		# Récupérer la partie correspondant à l'injection CSS s'il y
		# en a une.

		def get_props(value):
			props = ''
			if value != None:
				index = value.find(';')
				if index >= 0:
					props = value[index + 1:]
					value = value[:index]
			return value, props

		# Définir les propriétés à partir du truc principal.

		name = name[1:-1]
		props = ""
		if name == "b":
			self._bold = True
		elif name == "i":
			self._italic = True
		elif name == "u":
			self._underline = True
		elif name in ("s", "strike", "striked"):
			self._strike = True
		elif name == "color":
			value, props = get_props(value)
			self._color = get_color(value)
		elif name == "font":
			value, props = get_props(value)
			assert value in _fonts
			self._font = _fonts[value]
		elif name in ('size', 'big', 'small'):
			if name != 'size':
				value = name
			if value == 'big':
				self._size = _big_size
			elif value == 'small':
				self._size = _sml_size
			else:
				self._size = round(int(value) / 100.0, 2)
				assert 0 < self._size <= 3.0

			if self._size == 1.0:
				self._size = None
		elif name in _fonts:
			self._font = name
		else:
			self._color = get_color(name)

		# Gestion des injections CSS.

		for prop in props.split(';'):
			prop = prop.strip()
			if not prop:
				continue

			name, *value = prop.split(':')
			if not value:
				continue
			name = name.strip()
			value = ':'.join(value).strip()

			if name == 'size':
				unit = 'pt'
				if   value.endswith('pt'):
					value = value[:-2].rstrip()
				elif value.endswith('em'):
					unit = 'em'
					value = value[:-2].rstrip()

				if not value or \
					any(c != '0' for c in value[:-3]) or \
					any(not c in '0123456789' for c in value[-3:]):
					continue

				value = int(value[-3:])
				if unit == 'pt':
					value /= 12 # XXX: default em size

				if 0 < value <= 3.0:
					self._size = value

	def _get_css(self):
		""" Get the `style` CSS classes and properties for HTML output. """

		classes, props = [], []

		if not self.tweak('obsolete_tags', True):
			if self._bold:
				props.append('font-weight: bold')
			if self._italic:
				props.append('font-style: italic')
			if self._underline or self._strike:
				props.append('text-decoration:{}{}'.format(' underline' \
					if self._underline else '', ' line-through' \
					if self._strike else ''))

		if self._font:
			props.append('font-family: ' + self._font)

		if self._color:
			# `transparent` is at least considered as a special value,
			# or at most as an alias to `rgba(0,0,0,0)`.

			if self._color[3] == 0.0:
				props.append('color: transparent')
			else:
				# always append the #rgb color: it will be read by older
				# browsers if the `rgba()` function isn't supported.

				props.append('color: #%02X%02X%02X' % self._color[0:3])
				if self._color[3] < 1.0:
					props.append('color: rgba({}, {}, {}, {})' \
						.format(*self._color))

		if self._size:
			props.append('font-size: {}em'.format(self._size))

		return classes, props

	def begin_html(self):
		obsoletetags = self.tweak('obsolete_tags', True)

		cls, props = self._get_css()
		if cls or props:
			props = '<span{}{}>'.format(' class="{}"'.format(' '.join(cls)) \
				if cls else '', ' style="{}"'.format('; '.join(props)) \
				if props else '')
		else:
			props = ''

		return '' \
			+ ('', '<b>')[obsoletetags and self._bold] \
			+ ('', '<i>')[obsoletetags and self._italic] \
			+ ('', '<u>')[obsoletetags and self._underline] \
			+ ('', '<strike>')[obsoletetags and self._strike] \
			+ props

	def end_html(self):
		obsoletetags = self.tweak('obsolete_tags', True)

		return '' \
			+ ('', '</span>')[any(self._get_css())] \
			+ ('', '</strike>')[obsoletetags and self._strike] \
			+ ('', '</u>')[obsoletetags and self._underline] \
			+ ('', '</i>')[obsoletetags and self._italic] \
			+ ('', '</b>')[obsoletetags and self._bold]

# End of file.
